#define POSIX_C_SOURCE 200809L

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <errno.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include <CL/cl.h>

#define MAX_PLATFORM    3

struct ocl_data {
	cl_context ctx;
	cl_device_id devid;
	cl_command_queue queue;
};


static
void print_buildlog(FILE * fp, struct ocl_data* ocl, cl_program prg)
{
	size_t len;
	cl_int status;

	// check build error and build status first
	clGetProgramBuildInfo(prg, ocl->devid,
	                      CL_PROGRAM_BUILD_STATUS,
			      sizeof(cl_build_status), &status, NULL);

	// check build log
	clGetProgramBuildInfo(prg, ocl->devid, CL_PROGRAM_BUILD_LOG,
	                      0, NULL, &len);
	char log[len+1];
	clGetProgramBuildInfo(prg, ocl->devid, CL_PROGRAM_BUILD_LOG,
	                      len, log, NULL);
	log[len] = '\n';
	fprintf(fp, "Build status=%d, log:\n", status);
	fwrite(log, 1, sizeof(log), fp);
}


static
int find_ocl_device(cl_device_id * devid, cl_device_type type)
{
	cl_int err;
	cl_uint i, num_platforms, num_devices;
	cl_platform_id platform_ids[MAX_PLATFORM];

	err = clGetPlatformIDs(MAX_PLATFORM, platform_ids, &num_platforms);

	if (err != CL_SUCCESS) {
		fprintf(stderr, "GetPlatformIDs() failed: %i\n", err);
		return -1;
	}

	for (i = 0; i < num_platforms; i++) {
		err = clGetDeviceIDs(platform_ids[i], type, 1, devid,
		                     &num_devices);
		if (err != CL_SUCCESS)
			continue;

		// A device with the proper type has been found
		if (num_devices > 0)
			return 0;
	}

	fprintf(stderr, "Cannot find a matching OpenCL device.\n");
	return -1;
}


static
int init_ocl(struct ocl_data* ocl)
{
	cl_context ctx = NULL;
	cl_device_id devid = NULL;
	cl_platform_id pl;
	cl_command_queue queue = NULL;
	cl_int err = 0;
	char str[512];

	// reset all members in ocl
	ocl->ctx = NULL;
	ocl->devid = NULL;
	ocl->queue = NULL;

	// Find a suitable OpenCL device, create the associated context and
	// command queue
	if ( find_ocl_device(&devid, CL_DEVICE_TYPE_GPU)
	  || !(ctx = clCreateContext(NULL, 1, &devid, NULL, NULL, &err))
	  || !(queue = clCreateCommandQueue(ctx, devid, 0, &err)) ) {
		if (err)
			fprintf(stderr, "OpenCL init failed: %i\n", err);
		clReleaseContext(ctx);
		return -1;
	}

	// Display info about the selected device
	clGetDeviceInfo(devid, CL_DEVICE_PLATFORM, sizeof(pl), &pl, NULL);
	clGetPlatformInfo(pl, CL_PLATFORM_NAME, sizeof(str), str, NULL);
	fprintf(stderr, "OpenCL platform name: %s\n", str);
	clGetPlatformInfo(pl, CL_PLATFORM_VERSION, sizeof(str), str, NULL);
	fprintf(stderr, "OpenCL version: %s\n", str);
	clGetDeviceInfo(devid, CL_DEVICE_NAME, sizeof(str), str, NULL);
	fprintf(stderr, "OpenCL device: %s\n", str);

	ocl->ctx = ctx;
	ocl->devid = devid;
	ocl->queue = queue;
	return 0;
}


void cleanup_ocl(struct ocl_data* ocl)
{
	if (!ocl)
		return;

	clReleaseContext(ocl->ctx);
	clReleaseCommandQueue(ocl->queue);

	// reset all members in ocl
	ocl->ctx = NULL;
	ocl->devid = NULL;
	ocl->queue = NULL;
}


cl_program load_clprg(struct ocl_data* ocl, const char* filename)
{
	int fd;
	struct stat sb;
	char* clsrc = NULL;
	size_t srclen;
	cl_program prg = NULL;
	cl_int err = 0;

	// Open CL source file
	fd = open(filename, O_RDONLY);
	if (fd == -1) {
		fprintf(stderr, "Cannot open '%s': %s\n",
		                filename, strerror(errno));
		return NULL;
	}

	// Read CL source file and put it in clsrc and close file
	fstat(fd, &sb);
	srclen = sb.st_size;
	clsrc = malloc(srclen);
	read(fd, clsrc, srclen);
	close(fd);

	// Create and compile CL program
	fprintf(stderr, "going to compile CL prog\n");
	if (!(prg = clCreateProgramWithSource(ocl->ctx, 1, (const char**)&clsrc, &srclen, &err))
	  || (err = clBuildProgram(prg, 1, &ocl->devid, NULL, NULL, NULL) )) {
		fprintf(stderr, "Cannot create program (%i)\n", err);
		if (err == CL_BUILD_PROGRAM_FAILURE)
			print_buildlog(stderr, ocl, prg);

		if (prg) {
			clReleaseProgram(prg);
			prg = NULL;
		}
		goto exit;
	}

exit:
	free(clsrc);
	fprintf(stderr, "prog compiled\n");
	return prg;
}



int main(void)
{
	cl_program prg = NULL;
	int retval;
	struct ocl_data ocl;

	// Init OpenCL context, device and command queue
	if (init_ocl(&ocl))
		return EXIT_FAILURE;

	prg = load_clprg(&ocl, "test.cl");
	if (!prg) {
		retval = EXIT_FAILURE;
		goto exit;
	}

	retval = EXIT_SUCCESS;
	clReleaseProgram(prg);

exit:
	cleanup_ocl(&ocl);
	return retval;
}


